//=============================================================================
//
// Alazar Technologies Inc
//
// Copyright (c) 2018 Alazar Technologies Inc. All Rights Reserved.
// Unpublished - rights reserved under the Copyright laws of the
// United States And Canada.
//
// This product contains confidential information and trade secrets
// of Alazar Technologies Inc. Use, disclosure, or reproduction is
// prohibited without the prior express written permission of Alazar
// Technologies Inc
//
//=============================================================================

#ifndef UTILS_H
#define UTILS_H

#include "ats_error.h"
#include "ats_types.h"

#include "platform.h"

#ifndef NULL
#define NULL 0
#endif

#ifndef MIN
#define MIN(__x, __y) ((__x) < (__y) ? (__x) : (__y))
#endif

#ifndef MAX
#define MAX(__x, __y) ((__x) > (__y) ? (__x) : (__y))
#endif

#ifndef ARRAY_SIZE
#define ARRAY_SIZE(__a) (sizeof(__a) / sizeof((__a)[0]))
#endif

#ifndef ARRAY_END
#define ARRAY_END(__a) ((__a) + ARRAY_SIZE(__a))
#endif

#ifndef DIV_ROUND
#define DIV_ROUND(__x, __s) ((__x + (__s)-1) / (__s))
#endif

#ifndef OFFSET_OF
#define OFFSET_OF(__t, __m) ((size_t)(&((__t *)0)->__m))
#endif

#ifndef CONTAINER_OF
#define CONTAINER_OF(__p, __t, __m) (__t *)((char *)(__p)-OFFSET_OF(__t, __m))
#endif

#define KHZ(__f) (1000LL * (__f))
#define MHZ(__f) (1000000LL * (__f))

#define RET_ON_ERR(__c)                                                                            \
    do                                                                                             \
    {                                                                                              \
        int ret;                                                                                   \
        if ((ret = (__c)) < 0) return ret;                                                         \
    } while (0)

static inline int mask_to_id(unsigned long long value)
{
    int id;

    for (id = 0; (value > 0) && ((value & 1) == 0); value >>= 1, id++)
        ;

    return ((value > 0) ? id : -1);
}

static inline int popcount(unsigned int val)
{
#ifdef _WIN32
    int cnt;

    for (cnt = 0; val > 0; val >>= 1)
        cnt += (val & 1);

    return cnt;
#else
    return __builtin_popcount(val);
#endif
}

static inline bool is_power_of_two(unsigned int val) { return (popcount(val) == 1); }

/// @{
///
/// Checks if the input integer can be fit into an integer of a different type
/// of the given number of bits.
bool unsigned_fits_in_unsigned(unsigned long long value, int bitlen);
bool unsigned_fits_in_signed(unsigned long long value, int bitlen);
bool unsigned_fits_in_u32(long long value);
bool signed_fits_in_unsigned(long long value, int bitlen);
bool signed_fits_in_signed(long long value, int bitlen);
bool signed_fits_in_u32(long long value);
/// @}

/**
 * @brief Log buffer with line format "address: byte0 ... byte 15"
 *
 * @param level      Log level (ATS_LOG_ERROR, ATS_LOG_WARNING, ATS_LOG_INFO,
 * ATS_LOG_DEBUG)
 *
 * @param ident      Name of the buffer to log
 *
 * @param buffer     Data buffer to log
 *
 * @param size       Buffer size in bytes
 *
 */
void ats_log_buffer(int level, char *ident, unsigned char *buffer, int size);

/**
 * @brief Convert memory size Id from `MemorySizesPerChannel`
 *         to size in samples per channel.
 *
 * @param id         Memory size Id
 *
 */
long long memory_size_id_to_samples(int id);

/**
 * Rational number
 */
typedef struct rational_s
{
    long long num;
    long long den;
} rational_t;

/**
 * @Brief Reduces `rat` to lowest terms. A rational number is reduced when the
 * greatest common divisor of its numerator and denominator is one.
 *
 * @param rat      Rational number
 */
rational_t rational_fraction_reduce(rational_t rat);

/**
 * @Brief Constructs a rational number, reduced to lowest terms.
 *
 * @param num      Numerator
 * @param den      Denominator
 */
rational_t rational(long long num, long long den);

/**
 * @Brief Return the the converted integer value of a rational number.
 *
 * @param rat      Rational number to get value from
 */
long long rational_to_integer(rational_t rat);

/**
 * @Brief Returns `true` if the denominator of `rat` is non-null.
 *
 * @param rat      Rational rational object to check
 */
bool rational_is_valid(rational_t rat);

/**
 * @Brief Returns `true` if the rational number is equal to value.
 *
 * @param rat      Rational number to compare
 * @param val      Rational number to compare with
 */
bool rational_equal_to(rational_t rat, rational_t val);

/**
 * @Brief Returns `true` if the rational number is greater than value.
 *
 * @param rat      Rational number to compare
 * @param val      Rational number to compare with
 */
bool rational_greater_than(rational_t rat, rational_t val);

/**
 * @Brief Returns `true` if the rational number is less than value.
 *
 * @param rat      Rational number to compare
 * @param val      Rational number to compare with
 */
bool rational_less_than(rational_t rat, rational_t val);

/**
 * @Brief Returns the rational number resulting from the multiplication of two rational numbers.
 *
 * @param x        Rational number to multiplicate
 * @param y        Rational number multiplicator
 */
rational_t rational_mul(rational_t x, rational_t y);

/**
 * @Brief Returns the rational number resulting from the division of two rational numbers.
 *
 * @param x        Rational number to divide
 * @param y        Rational number divider
 */
rational_t rational_div(rational_t x, rational_t y);

/**
 * @Brief Returns the rational number resulting from the addition of two rational numbers.
 *
 * @param x        Rational number
 * @param y        Rational number to add
 */
rational_t rational_add(rational_t x, rational_t y);

/**
 * @Brief Returns the rational number resulting from the substraction of two rational numbers.
 *
 * @param x        Rational number
 * @param y        Rational number to substract
 */
rational_t rational_sub(rational_t x, rational_t y);

/**
 * @Brief Returns `true` if the rational number is in the range of min and max values.
 *
 * @param rat      Rational number to compare
 * @param min      Integer minimal value of range
 * @param max      Integer maximal value of range
 */
bool rational_in_integer_range(rational_t rat, long long min, long long max);

/**
 * @Brief Returns `true` if the rational number is equal to value.
 *
 * @param rat      Rational number to compare
 * @param val      Integer number to compare with
 */
bool rational_equal_to_integer(rational_t rat, long long val);

/**
 * @Brief Returns `true` if the rational number is greater than value.
 *
 * @param rat      Rational number to compare
 * @param val      Integer number to compare with
 */
bool rational_greater_than_integer(rational_t rat, long long val);

/**
 * @Brief Returns `true` if the rational number is less than value.
 *
 * @param rat      Rational number to compare
 * @param val      Integer number to compare with
 */
bool rational_less_than_integer(rational_t rat, long long val);

/**
 * @Brief Returns the rational number resulting from the multiplication of a rational number with an
 * integer.
 *
 * @param rat      Rational number to multiply
 * @param val      Integer to multiply with
 */
rational_t rational_mul_integer(rational_t rat, int val);

/**
 * @Brief Returns the rational number resulting from the division of a rational number with an
 * integer.
 *
 * @param rat      Rational number to divide
 * @param val      Integer to divide with
 */
rational_t rational_div_integer(rational_t rat, int val);

/**
 * @Brief Returns the rational number resulting from the modulo of a rational number with an
 * integer.
 *
 * @param rat      Rational number to modulo
 * @param val      Integer to modulo with
 */
rational_t rational_mod_integer(rational_t rat, int val);

/**
 * @brief Return true if rational can be converted to an integer value.
 *
 * @param rat      Rational number to check
 *
 */
bool rational_is_integer(rational_t rat);

/**
 * @brief Convert a sampling rate id to its frequency in Hz.
 *
 * @param rate     The sampling rate id (Member of ALAZAR_SAMPLE_RATES enum)
 *
 */
long long sampling_rate_id_to_hz(int rate);

/**
 * @brief Convert a sampling rate id to a rational number representing its frequency in Hz.
 *
 * @param rate     The sampling rate id (Member of ALAZAR_SAMPLE_RATES enum)
 *
 */
rational_t sampling_rate_id_to_rational(int rate);

#define assert(x)                                                                                  \
    do                                                                                             \
    {                                                                                              \
        if (!(x)) { ats_log_error("ASSERT error: %s failed", #x); }                                \
    } while (0)

#define rational_list(...)                                                                         \
    (rational_t[]) { __VA_ARGS__, rational(0, 0) }

typedef struct
{
    rational_t clock;
    int ratio;
} clock_decimation_t;

typedef bool (*clock_supported_decimation_t)(int);

rc_t clock_decimation(clock_decimation_t *decimation,
                      clock_supported_decimation_t supported,
                      rational_t *clocks,
                      rational_t rate);

#define TIME_MEASUREMENT(__x, __s)                                                                 \
    do                                                                                             \
    {                                                                                              \
        static int64_t __update = 0;                                                               \
        static size_t __count = 0;                                                                 \
        int64_t __start = ats_get_utime();                                                         \
        __x;                                                                                       \
        __update += (ats_get_utime() - __start);                                                   \
        __count++;                                                                                 \
        if (__count < __s) break;                                                                  \
        ats_log_info("%s() @%u time=%lluus\n", __FUNCTION__, __LINE__, __update / __count);        \
        __update = 0;                                                                              \
        __count = 0;                                                                               \
    } while (0)

#endif // UTILS_H
